# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Any

import yaml
from mutenix.teams_messages import MeetingAction
from pydantic import BaseModel

_logger = logging.getLogger(__name__)

CONFIG_FILENAME = "mutenix.yaml"


class ActionEnum(str, Enum):
    ACTIVATE_TEAMS = "activate-teams"
    CMD = "cmd"
    WEBHOOK = "webhook"


class LedStatusSource(str, Enum):
    TEAMS = "teams"
    CMD = "cmd"


class LedColor(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    WHITE = "white"
    BLACK = "black"
    YELLOW = "yellow"
    CYAN = "cyan"
    MAGENTA = "magenta"
    ORANGE = "orange"
    PURPLE = "purple"


class TeamsState(str, Enum):
    MUTED = "is-muted"
    HAND_RAISED = "is-hand-raised"
    IN_MEETING = "is-in-meeting"
    RECORDING_ON = "is-recording-on"
    BACKGROUND_BLURRED = "is-background-blurred"
    SHARING = "is-sharing"
    UNREAD_MESSAGES = "has-unread-messages"
    VIDEO_ON = "is-video-on"


class ButtonAction(BaseModel):
    button_id: int
    action: MeetingAction | ActionEnum
    extra: str | dict[str, Any] | None = None


class LedStatus(BaseModel):
    button_id: int
    source: LedStatusSource
    extra: TeamsState | str | None = None
    color_on: LedColor | None = None
    color_off: LedColor | None = None
    read_result: bool = False
    interval: float = 0.0


class VirtualKeypadConfig(BaseModel):
    bind_address: str = "127.0.0.1"
    bind_port: int = 12909


class DeviceInfo(BaseModel):
    vendor_id: int | None = None
    product_id: int | None = None
    serial_number: str | None = None


class Config(BaseModel):
    actions: list[ButtonAction]
    double_tap_action: list[ButtonAction] = []
    leds: list[LedStatus] = []
    teams_token: str | None = None
    file_path: str | None = None
    virtual_keypad: VirtualKeypadConfig = VirtualKeypadConfig()
    auto_update: bool = True
    device_identifications: list[DeviceInfo] = []


def create_default_config() -> Config:
    return Config(
        actions=[
            ButtonAction(button_id=1, action=MeetingAction.ToggleMute),
            ButtonAction(button_id=2, action=MeetingAction.ToggleHand),
            ButtonAction(button_id=3, action=ActionEnum.ACTIVATE_TEAMS),
            ButtonAction(button_id=4, action=MeetingAction.React, extra="like"),
            ButtonAction(button_id=5, action=MeetingAction.LeaveCall),
            ButtonAction(button_id=6, action=MeetingAction.ToggleMute),
            ButtonAction(button_id=7, action=MeetingAction.ToggleHand),
            ButtonAction(button_id=8, action=ActionEnum.ACTIVATE_TEAMS),
            ButtonAction(button_id=9, action=MeetingAction.React, extra="like"),
            ButtonAction(button_id=10, action=MeetingAction.LeaveCall),
        ],
        double_tap_action=[
            ButtonAction(button_id=3, action=MeetingAction.ToggleVideo),
            ButtonAction(button_id=8, action=MeetingAction.ToggleVideo),
        ],
        leds=[
            LedStatus(
                button_id=1,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.MUTED,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=2,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.HAND_RAISED,
                color_on=LedColor.YELLOW,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=3,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.VIDEO_ON,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=5,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.IN_MEETING,
                color_on=LedColor.GREEN,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=6,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.MUTED,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=7,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.HAND_RAISED,
                color_on=LedColor.YELLOW,
                color_off=LedColor.BLACK,
            ),
            LedStatus(
                button_id=8,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.VIDEO_ON,
                color_on=LedColor.RED,
                color_off=LedColor.GREEN,
            ),
            LedStatus(
                button_id=10,
                source=LedStatusSource.TEAMS,
                extra=TeamsState.IN_MEETING,
                color_on=LedColor.GREEN,
                color_off=LedColor.BLACK,
            ),
        ],
        teams_token=None,
        virtual_keypad=VirtualKeypadConfig(bind_address="127.0.0.1", bind_port=12909),
    )


def find_config_file() -> Path:
    file_path = Path(CONFIG_FILENAME)
    home_config_path = (
        Path.home() / os.environ.get("XDG_CONFIG_HOME", ".config") / CONFIG_FILENAME
    )

    if not file_path.exists() and home_config_path.exists():
        file_path = home_config_path

    return file_path


def load_config(file_path: Path | None = None) -> Config:
    if file_path is None:
        file_path = find_config_file()

    try:
        _logger.info("Loading config from file: %s", file_path)
        with open(file_path, "r") as file:
            config_data = yaml.safe_load(file)
        if config_data is None:
            raise yaml.YAMLError("No data in file")
    except (FileNotFoundError, yaml.YAMLError, IOError) as e:
        print(e)

        config = create_default_config()
        config.file_path = str(file_path)
        save_config(config)
        return config

    config_data["file_path"] = str(file_path)
    return Config(**config_data)


def save_config(config: Config, file_path: Path | str | None = None):
    if file_path is None:
        if config.file_path is None:
            raise ValueError("No file path provided")  # pragma: no cover

        file_path = config.file_path

    config.file_path = None
    try:
        with open(file_path, "w") as file:
            yaml.dump(config.model_dump(mode="json"), file)
    except (FileNotFoundError, yaml.YAMLError, IOError):
        _logger.error("Failed to write config to file: %s", file_path)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import tracemalloc
from unittest.mock import ANY
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import Ping
from mutenix.hid_commands import PrepareUpdate
from mutenix.hid_device import HidDevice

tracemalloc.start()


@pytest.fixture
def hid_device():
    with patch("hid.device") as MockHidDevice:
        MockHidDevice.return_value = Mock()
        return HidDevice()


@pytest.mark.asyncio
async def test_send_msg(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    assert not future.done()
    assert not hid_device._send_buffer.empty()


@pytest.mark.asyncio
async def test_register_callback(hid_device):
    callback = Mock()
    hid_device.register_callback(callback)
    assert callback in hid_device._callbacks


@pytest.mark.asyncio
async def test_ping(hid_device):
    with patch.object(hid_device, "send_msg", return_value=Mock()):
        with patch("asyncio.sleep", side_effect=asyncio.CancelledError):
            with pytest.raises(asyncio.CancelledError):
                await hid_device._ping()


@pytest.mark.asyncio
async def test_write_success(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(hid_device, "_send_report", return_value=1):
        await hid_device._write()
    assert future.done()
    assert future.result() == 1


@pytest.mark.asyncio
async def test_write_failure(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(hid_device, "_send_report", return_value=-1):
        await hid_device._write()
    assert future.done()
    assert future.exception() is not None


@pytest.mark.asyncio
async def test_write_device_disconnected(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(
        hid_device,
        "_send_report",
        side_effect=OSError("Device disconnected"),
    ):
        await hid_device._write()
    assert not future.done()


@pytest.mark.asyncio
async def test_write_value_error(hid_device):
    msg = HidOutputMessage()
    future = hid_device.send_msg(msg)
    with patch.object(
        hid_device,
        "_send_report",
        side_effect=ValueError("Invalid message"),
    ):
        await hid_device._write()
    assert not future.done()


@pytest.mark.asyncio
async def test_read_success(hid_device):
    data = bytes([0x01, 0x02, 0x03])
    future = asyncio.get_event_loop().create_future()

    async def callback(msg):
        future.set_result(msg)

    hid_device._device = Mock()
    hid_device._device.read.return_value = data

    hid_device._callbacks.append(callback)
    with patch(
        "mutenix.hid_commands.HidInputMessage.from_buffer",
        return_value=Mock(),
    ):
        await hid_device._read()
    await future


@pytest.mark.asyncio
async def test_read_device_disconnected(hid_device):
    hid_device._wait_for_device = AsyncMock()
    hid_device._device = Mock()
    hid_device._device.read.side_effect = OSError("Device disconnected")

    with patch("mutenix.hid_device._logger.error") as mock_logger:
        await hid_device._read()
        mock_logger.assert_called_with("Device disconnected: %s", ANY)

    assert hid_device._wait_for_device.called


@pytest.mark.asyncio
async def test_read_value_error(hid_device):
    hid_device._device = Mock()
    hid_device._device.read.side_effect = ValueError("Invalid message")
    with patch("mutenix.hid_device._logger.error") as mock_logger:
        await hid_device._read()
        mock_logger.assert_called_with(
            "Error reading message: %s",
            ANY,
        )


class TypeMatcher:
    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


@pytest.mark.asyncio
async def test_ping_sends_ping_message(hid_device):
    hid_device._last_communication = asyncio.get_event_loop().time() - 5
    with patch.object(hid_device, "send_msg") as mock_send_msg:
        await hid_device._ping()
        mock_send_msg.assert_called_once_with(TypeMatcher(Ping))


@pytest.mark.asyncio
async def test_ping_resets_last_communication(hid_device):
    initial_time = asyncio.get_event_loop().time()
    with patch("asyncio.get_event_loop", return_value=asyncio.get_event_loop()):
        with patch.object(hid_device, "send_msg"):
            await hid_device._ping()
            assert hid_device._last_ping_time >= initial_time


class RoundAboutMatcher:
    def __init__(self, expected_value, offset):
        self.expected_value = expected_value
        self.offset = offset

    def __eq__(self, other):
        return abs(other - self.expected_value) < self.offset


@pytest.mark.asyncio
async def test_ping_waits_before_sending(hid_device):
    hid_device._last_ping_time = asyncio.get_event_loop().time()
    with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
        with patch.object(hid_device, "send_msg"):
            await hid_device._ping()
            mock_sleep.assert_called_with(RoundAboutMatcher(4.5, 0.2))


# test


@pytest.mark.asyncio
async def test_send_report_success(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.return_value = 1
    result = hid_device._send_report(msg)
    assert result == 1
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_send_report_failure(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.return_value = -1
    result = hid_device._send_report(msg)
    assert result == -1
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_send_report_exception(hid_device):
    msg = PrepareUpdate()

    hid_device._device = Mock()
    hid_device._device.write.side_effect = OSError("Device error")
    with pytest.raises(OSError, match="Device error"):
        hid_device._send_report(msg)
    hid_device._device.write.assert_called_once_with(
        bytes([msg.REPORT_ID]) + msg.to_buffer(),
    )


@pytest.mark.asyncio
async def test_unregister_callback(hid_device: HidDevice):
    # Create a HidDevice instance
    device = HidDevice()

    # Create a mock callback
    callback = Mock()

    # Register the callback
    device.register_callback(callback)
    assert callback in device._callbacks

    # Unregister the callback
    device.unregister_callback(callback)
    assert callback not in device._callbacks

    # Try to unregister the callback again (should not raise an error)
    device.unregister_callback(callback)
    assert callback not in device._callbacks


@pytest.mark.asyncio
async def test_invoke_callbacks_sync(hid_device):
    callback = Mock()
    hid_device.register_callback(callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    callback.assert_called_once_with(msg)


@pytest.mark.asyncio
async def test_invoke_callbacks_async(hid_device):
    async def async_callback(msg):
        async_callback.called = True

    async_callback.called = False
    hid_device.register_callback(async_callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    await asyncio.sleep(0)  # Allow the async callback to be called
    assert async_callback.called


@pytest.mark.asyncio
async def test_invoke_callbacks_mixed(hid_device):
    callback = Mock()

    async def async_callback(msg):
        async_callback.called = True

    async_callback.called = False
    hid_device.register_callback(callback)
    hid_device.register_callback(async_callback)
    msg = Mock()
    hid_device._invoke_callbacks(msg)
    callback.assert_called_once_with(msg)
    await asyncio.sleep(0)  # Allow the async callback to be called
    assert async_callback.called


@pytest.mark.asyncio
async def test_search_for_device_success(hid_device):
    device_info = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
    }
    with patch("hid.device"):
        with patch("hid.enumerate", return_value=[device_info]):
            device = await hid_device._search_for_device()
            assert device is not None


@pytest.mark.asyncio
async def test_search_for_device_prefer_bluetooth(hid_device):
    device_info_bt = {
        "product_id": 0,
        "vendor_id": 0,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
    }
    device_info_usb = {
        "product_id": 0x1234,
        "vendor_id": 0x5678,
        "serial_number": "122345",
        "product_string": "Mutenix Macropad",
    }
    with patch("hid.device") as MockHidDevice:
        with patch("hid.enumerate", return_value=[device_info_bt, device_info_usb]):
            m = Mock()
            MockHidDevice.return_value = m
            device = await hid_device._search_for_device()
            assert device is not None
            m.open.assert_called_with(product_id=0, vendor_id=0, serial_number="122345")


@pytest.mark.asyncio
async def test_search_for_device_no_device_found(hid_device):
    with patch("hid.enumerate", return_value=[]):
        device = await hid_device._search_for_device()
        assert device is None


@pytest.mark.asyncio
async def test_search_for_device_exception(hid_device):
    with patch("hid.enumerate", side_effect=Exception("Error")):
        with patch("mutenix.hid_device._logger.debug") as mock_logger:
            device = await hid_device._search_for_device()
            assert device is None
            mock_logger.assert_called_with("Failed to get device: %s", ANY)


@pytest.mark.asyncio
async def test_wait_for_device_success(hid_device):
    with patch.object(
        hid_device,
        "_search_for_device_loop",
        new_callable=AsyncMock,
    ) as mock_search:
        mock_search.return_value = Mock()
        await hid_device._wait_for_device()
        assert mock_search.called
        assert hid_device._device is not None


@pytest.mark.asyncio
async def test_wait_for_device_failure(hid_device):
    with patch.object(
        hid_device,
        "_search_for_device_loop",
        new_callable=AsyncMock,
    ) as mock_search:
        mock_search.return_value = None
        await hid_device._wait_for_device()
        assert mock_search.called
        assert hid_device._device is None



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import time
from unittest.mock import patch

import pytest
from mutenix.utils import block_parallel
from mutenix.utils import run_till_some_loop


@pytest.mark.asyncio
async def test_block_parallel_allows_single_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_block_parallel_blocks_parallel_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()

    async def call_method():
        return await test_instance.test_method()

    task1 = asyncio.create_task(call_method())
    task2 = asyncio.create_task(call_method())

    results = await asyncio.gather(task1, task2)
    assert results == ["done", None]


@pytest.mark.asyncio
async def test_block_parallel_resets_after_execution():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    await test_instance.test_method()
    result = await test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_block_parallel_logs_debug_messages():
    class TestClass:
        @block_parallel
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()

    with patch("mutenix.utils._logger.debug") as mock_debug:
        await test_instance.test_method()
        mock_debug.assert_called_with("block_parallel %s %s", "test_method", False)


@pytest.mark.asyncio
async def test_run_till_some_loop_async():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        async def test_method(self):
            await asyncio.sleep(0.1)
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


def test_run_till_some_loop_sync():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        def test_method(self):
            time.sleep(0.1)
            return "done"

    test_instance = TestClass()
    result = test_instance.test_method()
    assert result == "done"


@pytest.mark.asyncio
async def test_run_till_some_loop_async_stops_on_condition():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        async def test_method(self):
            await asyncio.sleep(0.1)
            self._run = False
            return "done"

    test_instance = TestClass()
    result = await test_instance.test_method()
    assert result == "done"


def test_run_till_some_loop_sync_stops_on_condition():
    class TestClass:
        _run = True

        @run_till_some_loop(sleep_time=0.1)
        def test_method(self):
            time.sleep(0.1)
            self._run = False
            return "done"

    test_instance = TestClass()
    result = test_instance.test_method()
    assert result == "done"



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger <matthias@bilger.info>
import asyncio
import json
import logging
import pathlib
from typing import Callable

import jinja2
import markdown
from aiohttp import web
from aiohttp_jinja2 import render_template
from aiohttp_jinja2 import setup as jinja2_setup
from mutenix.hid_commands import HidOutputMessage
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status

HOST = "127.0.0.1"
PORT = 12909

_logger = logging.getLogger(__name__)


class UnsupportedMessageTypeError(Exception):
    """Exception raised for unsupported message types in VirtualMacropad."""

    pass


class VirtualMacropad:
    """A virtual representation of the Macropad for testing or playing around."""

    icons: list[dict[str, str]] = [
        {
            "src": "/favicon/32",
            "sizes": "32x32",
            "type": "image/png",
        },
        {
            "src": "/favicon/16",
            "sizes": "16x16",
            "type": "image/png",
        },
        {
            "src": "/favicon/apple_touch",
            "sizes": "180x180",
            "type": "image/png",
        },
    ]

    def __init__(self, host: str = HOST, port: int = PORT):
        self.host = host
        self.port = port
        self._callbacks: list[Callable[[HidOutputMessage], asyncio.Future]] = []
        self.app = web.Application()
        self.app.router.add_static(
            "/static/",
            path=str(pathlib.Path(__file__).parent / "static"),
            name="static",
        )
        self.app.router.add_route("GET", "/favicon/{filename}", self.favicon)
        self.app.router.add_route("GET", "/favicon.svg", self.favicon_svg)
        self.app.router.add_route("GET", "/site.webmanifest", self.serve_manifest)
        self.app.router.add_route("GET", "/help", self.help)
        self.app.router.add_route("GET", "/about", self.about)
        self.app.add_routes(
            [
                web.get("/", self.index),
                web.post("/button", self.button_handler),
                web.get("/ws", self.websocket_handler),
            ],
        )
        jinja2_setup(self.app, loader=jinja2.PackageLoader("mutenix", "templates"))
        self._websockets: set[web.WebSocketResponse] = set()
        self._led_status: dict[int, str] = {}
        self._led_status_lock = asyncio.Lock()

    def register_callback(self, callback: Callable[[HidOutputMessage], asyncio.Future]):
        self._callbacks.append(callback)

    async def index(self, request: web.Request):
        return render_template("index.html", request, {})

    async def button_handler(self, request: web.Request):
        data = await request.json()
        await self._handle_msg(Status.trigger_button(data.get("button")))
        return web.Response(status=200)

    async def favicon(self, request: web.Request):
        filename = request.match_info["filename"]
        for icon in self.icons:
            if icon["src"].endswith(filename):
                icon_path = (
                    pathlib.Path(__file__).parent
                    / "assets"
                    / f"icon_active_{filename}.png"
                )
                break
        else:
            raise web.HTTPNotFound()
        return web.FileResponse(icon_path)

    async def favicon_svg(self, request: web.Request):
        return web.FileResponse(
            pathlib.Path(__file__).parent
            / "assets"
            / "mutenix_logo_finalicon_active.svg",
        )

    async def serve_manifest(self, request: web.Request):
        manifest = {
            "name": "Mutenix Virtual Macropad",
            "short_name": "Mutenix",
            "icons": self.icons,
            "start_url": "/",
            "display": "standalone",
        }
        return web.json_response(manifest)

    async def help(self, request: web.Request):
        return render_template("help.html", request, {})

    async def about(self, request: web.Request):
        readme_path = pathlib.Path(__file__).parent.parent.parent / "README.md"
        license_path = pathlib.Path(__file__).parent.parent.parent / "LICENSE"
        with open(readme_path, "r", encoding="utf-8") as f:
            readme_content = f.read()
        with open(license_path, "r", encoding="utf-8") as f:
            license_content = f.read()
        html_readme_content = markdown.markdown(readme_content)
        html_license_content = markdown.markdown(license_content)
        context = {
            "readme_content": html_readme_content,
            "license_content": html_license_content,
        }
        return render_template("about.html", request, context)

    async def _handle_msg(self, msg: HidOutputMessage):
        for callback in self._callbacks:
            await callback(msg)

    async def handle_state_request(self, ws):
        async with self._led_status_lock:
            for i, color in self._led_status.items():
                if color:
                    await ws.send_json({"button": i, "color": color})

    async def websocket_handler(self, request: web.Request):
        ws = web.WebSocketResponse()
        await ws.prepare(request)
        self._websockets.add(ws)
        async for msg in ws:
            if msg.type == web.WSMsgType.TEXT:
                data = json.loads(msg.data)
                match data["command"]:
                    case "button":
                        await self._handle_msg(
                            Status.trigger_button(data.get("button")),
                        )
                    case "state_request":
                        await self.handle_state_request(ws)
                    case _:
                        _logger.info("Unknown message: %s", data)
                        await ws.send_json({"error": "unknown command"})
            else:
                _logger.info("Unknown message: %s", msg)
                await ws.send_json({"error": "unknown message"})
        self._websockets.remove(ws)
        return ws

    @staticmethod
    async def _send_json_safe(ws, data):
        try:
            await ws.send_json(data)
        except Exception as e:
            _logger.error("Error sending LED status: %s to websocket %s", e, ws)

    def _send_led_status(self, button: int, color: str):
        for ws in self._websockets:
            asyncio.create_task(
                self._send_json_safe(ws, {"button": button, "color": color}),
            )

    async def send_msg(self, msg: HidOutputMessage):
        if isinstance(msg, SetLed):
            color = msg.color.name.lower()
            async with self._led_status_lock:
                self._led_status[msg.id] = color
            self._send_led_status(msg.id, color)
        else:
            raise UnsupportedMessageTypeError("Unsupported message type")
        _logger.debug("Sent message: %s", msg)

    async def process(self):
        runner = web.AppRunner(self.app, access_log=None)
        await runner.setup()
        site = web.TCPSite(runner, self.host, self.port)
        await site.start()
        _logger.info("VirtualMacropad running at http://%s:%s", self.host, self.port)

    async def stop(self):
        await self.app.shutdown()
        await self.app.cleanup()
        _logger.info("VirtualMacropad stopped")



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import pytest
from mutenix.hid_commands import HidInCommands
from mutenix.hid_commands import HidInputMessage
from mutenix.hid_commands import HidOutCommands
from mutenix.hid_commands import LedColor
from mutenix.hid_commands import Reset
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.hid_commands import StatusRequest
from mutenix.hid_commands import VersionInfo


def test_status():
    buffer = bytes([1, 1, 0, 0, 1])
    status = Status(buffer)
    assert status.button == 1
    assert status.triggered is True
    assert status.doubletap is False
    assert status.pressed is False
    assert status.released is True


def test_version_info():
    buffer = bytes([1, 0, 0, 3])
    version_info = VersionInfo(buffer)
    assert version_info.version == "1.0.0"
    assert version_info.type.name == "FIVE_BUTTON_USB"


def test_reset():
    reset = Reset()
    buffer = reset.to_buffer()
    assert buffer[:-1] == bytes([HidOutCommands.RESET, 0, 0, 0, 0, 0, 0, 0])[:-1]


def test_set_led():
    led = SetLed(1, LedColor.RED)
    buffer = led.to_buffer()
    assert (
        buffer[:-1]
        == bytes([HidOutCommands.SET_LED, 1, 0x00, 0x0A, 0x00, 0x00, 0, 0])[:-1]
    )


def test_from_buffer_version_info():
    buffer = bytes([0, HidInCommands.VERSION_INFO, 1, 2, 3, 4, 5, 6])
    message = HidInputMessage.from_buffer(buffer)
    assert isinstance(message, VersionInfo)
    assert message.buffer == buffer[2:8]


def test_from_buffer_status():
    buffer = bytes([0, HidInCommands.STATUS, 1, 0, 0, 1, 0, 0])
    message = HidInputMessage.from_buffer(buffer)
    assert isinstance(message, Status)
    assert message.buffer == buffer[2:8]


def test_from_buffer_status_request():
    buffer = bytes([0, HidInCommands.STATUS_REQUEST.value, 0, 0, 0, 0, 0, 0])
    message = HidInputMessage.from_buffer(buffer)
    assert isinstance(message, StatusRequest)


def test_from_buffer_not_implemented():
    buffer = bytes([0, 0xFF, 0, 0, 0, 0, 0, 0])
    with pytest.raises(NotImplementedError):
        HidInputMessage.from_buffer(buffer)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Matthias Bilger matthias@bilger.info
from __future__ import annotations

import asyncio
import pathlib
import time
from collections import defaultdict
from unittest.mock import ANY
from unittest.mock import AsyncMock
from unittest.mock import Mock
from unittest.mock import mock_open
from unittest.mock import patch

import pytest
from mutenix.config import create_default_config
from mutenix.config import LedStatusSource
from mutenix.hid_commands import LedColor
from mutenix.hid_commands import SetLed
from mutenix.hid_commands import Status
from mutenix.hid_commands import VersionInfo
from mutenix.macropad import Macropad
from mutenix.teams_messages import ClientMessageParameter
from mutenix.teams_messages import ClientMessageParameterType
from mutenix.teams_messages import MeetingAction
from mutenix.teams_messages import MeetingPermissions
from mutenix.teams_messages import MeetingState
from mutenix.teams_messages import MeetingUpdate
from mutenix.teams_messages import ServerMessage


@pytest.fixture
def macropad():
    with (
        patch("mutenix.macropad.HidDevice") as MockHidDevice,
        patch("mutenix.macropad.WebSocketClient") as MockWebSocketClient,
        patch("mutenix.macropad.VirtualMacropad") as MockVirtualMacropad,
        patch("mutenix.config.load_config") as MockLoadConfigFile,
        patch("mutenix.config.save_config"),
    ):
        MockHidDevice.return_value = Mock()
        MockWebSocketClient.return_value = Mock()
        MockVirtualMacropad.return_value = Mock()
        MockLoadConfigFile.return_value = create_default_config()
        return Macropad(pathlib.Path(__file__).parent / "mutenix.yaml")


@pytest.mark.asyncio
async def test_hid_callback_status(macropad):
    msg = Status(bytes([1, 1, 0, 0, 1]))
    macropad._websocket.send_message = AsyncMock()
    await macropad._hid_callback(msg)
    macropad._websocket.send_message.assert_called_once()
    assert (
        macropad._websocket.send_message.call_args[0][0].action
        == MeetingAction.ToggleMute
    )


@pytest.mark.asyncio
async def test_hid_callback_version_info(macropad):
    msg = VersionInfo(bytes([1, 0, 0, 2]))
    macropad._version_seen = None
    with patch(
        "mutenix.macropad.check_for_device_update",
    ) as mock_check_for_device_update:
        await macropad._hid_callback(msg)
        mock_check_for_device_update.assert_called_once_with(ANY, msg)


@pytest.mark.asyncio
async def test_hid_callback_version_info_only_once(macropad):
    msg = VersionInfo(bytes([1, 0, 0, 2]))
    macropad._version_seen = None
    await macropad._hid_callback(msg)
    with patch(
        "mutenix.macropad.check_for_device_update",
    ) as mock_check_for_device_update:
        await macropad._hid_callback(msg)
        mock_check_for_device_update.assert_not_called()


@pytest.mark.asyncio
async def test_teams_callback_token_refresh(macropad):
    msg = ServerMessage(tokenRefresh="new_token")
    macropad._current_state = None
    with patch("builtins.open", mock_open()) as mock_file:
        await macropad._teams_callback(msg)
        mock_file.assert_called_once()
        assert mock_file.call_args[0][0].endswith("mutenix.yaml")
        mock_file().write.assert_any_call("new_token")


@pytest.mark.asyncio
async def test_teams_callback_token_refresh_save_failed(macropad):
    msg = ServerMessage(tokenRefresh="new_token")
    macropad._current_state = None
    with (
        patch("builtins.open", mock_open()) as mock_file,
        patch("mutenix.macropad._logger.error"),
    ):
        mock_file().write.side_effect = IOError
        await macropad._teams_callback(msg)
        mock_file.assert_called()
        assert mock_file.call_args[0][0].endswith("mutenix.yaml")
        # mock_logger_error.assert_called_once()


@pytest.mark.asyncio
async def test_update_device_status(macropad):
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=True,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )

    def send_msg(msg):
        future = asyncio.get_event_loop().create_future()
        future.set_result(None)
        assert isinstance(msg, SetLed)
        return future

    macropad._device.send_msg = Mock(side_effect=send_msg)
    macropad._virtual_macropad.send_msg = Mock()
    await macropad._update_device_status()
    assert macropad._device.send_msg.call_count == 8
    assert macropad._virtual_macropad.send_msg.call_count == 8


@pytest.mark.asyncio
async def test_update_device_status_not_in_meeting(macropad):
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=False,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=False,
            ),
            meetingPermissions=MeetingPermissions(canLeave=False),
        ),
    )

    def send_msg(msg):
        future = asyncio.get_event_loop().create_future()
        future.set_result(None)
        assert isinstance(msg, SetLed)
        return future

    macropad._device.send_msg = Mock(side_effect=send_msg)
    macropad._virtual_macropad.send_msg = AsyncMock()
    await macropad._update_device_status()
    assert macropad._device.send_msg.call_count == 8
    assert macropad._virtual_macropad.send_msg.call_count == 8


@pytest.mark.parametrize(
    "msg_bytes, expected_action, should_call",
    [
        (bytes([1, 1, 0, 0, 1]), MeetingAction.ToggleMute, True),
        (bytes([2, 1, 0, 0, 1]), MeetingAction.ToggleHand, True),
        (bytes([4, 1, 0, 0, 1]), MeetingAction.React, True),
        (bytes([5, 1, 0, 0, 1]), MeetingAction.LeaveCall, True),
        (bytes([1, 0, 0, 0, 1]), None, False),
        (bytes([2, 0, 0, 0, 1]), None, False),
        (bytes([4, 0, 0, 0, 1]), None, False),
        (bytes([5, 0, 0, 0, 1]), None, False),
        (bytes([1, 1, 0, 0, 0]), None, False),
        (bytes([2, 1, 0, 0, 0]), None, False),
        (bytes([4, 1, 0, 0, 0]), None, False),
        (bytes([5, 1, 0, 0, 0]), None, False),
    ],
)
@pytest.mark.asyncio
async def test_hid_callback_parametrized(
    macropad,
    msg_bytes,
    expected_action: MeetingAction,
    should_call,
):
    msg = Status(msg_bytes)
    macropad._websocket.send_message = AsyncMock()

    await macropad._hid_callback(msg)
    if should_call:
        macropad._websocket.send_message.assert_called_once()
        if expected_action:
            assert (
                macropad._websocket.send_message.call_args[0][0].action.name
                == expected_action.name
            )
    else:
        macropad._websocket.send_message.assert_not_called()


@pytest.mark.asyncio
async def test_hid_callback_invalid_button(macropad):
    msg = Status([11, 1, 0, 0, 1])

    await macropad._hid_callback(msg)
    macropad._websocket.send_message.assert_not_called()


class IdentifierWithoutToken:
    def __eq__(self, value):
        if value.token:
            return False
        return True


@pytest.mark.asyncio
async def test_setup_without_existing_token():
    with patch("builtins.open", mock_open(read_data="")) as mock_file:
        mock_file.side_effect = FileNotFoundError
        with patch("mutenix.macropad.HidDevice"):
            with patch("mutenix.macropad.WebSocketClient") as MockWebSocketClient:
                with patch("mutenix.macropad.VirtualMacropad"):
                    macropad = Macropad()
                    macropad._setup()
                    MockWebSocketClient.assert_called_with(
                        ANY,
                        IdentifierWithoutToken(),
                    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "status, expected_action, expected_parameters",
    [
        (Status(bytes([1, 1, 0, 0, 1])), MeetingAction.ToggleMute, None),
        (Status(bytes([2, 1, 0, 0, 1])), MeetingAction.ToggleHand, None),
        # bring_teams_to_foreground
        (Status(bytes([3, 1, 0, 0, 1])), None, None),
        (Status(bytes([3, 1, 1, 0, 1])), MeetingAction.ToggleVideo, None),  # doubletap
        (
            Status(bytes([4, 1, 0, 0, 1])),
            MeetingAction.React,
            ClientMessageParameter(type_=ClientMessageParameterType.ReactLike),
        ),
        (Status(bytes([5, 1, 0, 0, 1])), MeetingAction.LeaveCall, None),
        (Status(bytes([1, 0, 0, 0, 1])), None, None),
    ],
)
async def test_send_status(macropad, status, expected_action, expected_parameters):
    macropad._websocket.send_message = AsyncMock()
    with patch(
        "mutenix.macropad.bring_teams_to_foreground",
    ) as mock_bring_teams_to_foreground:
        await macropad._send_status(status)
        if expected_action:
            macropad._websocket.send_message.assert_called_once()
            client_message = macropad._websocket.send_message.call_args[0][0]
            assert client_message.action == expected_action
            if expected_parameters:
                assert client_message.parameters == expected_parameters
        else:
            macropad._websocket.send_message.assert_not_called()
            if status.button == 3 and status.triggered and status.released:
                mock_bring_teams_to_foreground.assert_called_once()
            else:
                mock_bring_teams_to_foreground.assert_not_called()


@pytest.mark.asyncio
async def test_process(macropad):
    macropad._device.process = AsyncMock()
    macropad._websocket.process = AsyncMock()
    macropad._virtual_macropad.process = AsyncMock()

    async def stop_macropad():
        await asyncio.sleep(0.2)
        await macropad.stop()

    asyncio.create_task(stop_macropad())

    await macropad.process()

    macropad._device.process.assert_called_once()
    macropad._websocket.process.assert_called_once()
    macropad._virtual_macropad.process.assert_called_once()


@pytest.mark.asyncio
async def test_process_with_exception(macropad):
    macropad._device.process = AsyncMock(side_effect=Exception("Device error"))
    macropad._websocket.process = AsyncMock(return_value=None)
    macropad._virtual_macropad.process = AsyncMock(return_value=None)

    macropad._check_status = AsyncMock()

    with patch("mutenix.macropad._logger.error") as mock_logger_error:
        await macropad.process()
        mock_logger_error.assert_called_with("Error in Macropad process: %s", ANY)

    macropad._device.process.assert_called_once()
    macropad._websocket.process.assert_called_once()
    macropad._virtual_macropad.process.assert_called_once()


@pytest.mark.asyncio
async def test_manual_update_success(macropad):
    update_file = "update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open(read_data=b"update_data")) as mock_file:
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_called_once_with(
                macropad._device.raw,
                mock_file(),
            )


@pytest.mark.asyncio
async def test_manual_update_file_not_found(macropad):
    update_file = "non_existent_update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open()) as mock_file:
        mock_file.side_effect = FileNotFoundError
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            with pytest.raises(FileNotFoundError):
                await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_not_called()


@pytest.mark.asyncio
async def test_manual_update_io_error(macropad):
    update_file = "update.bin"
    macropad._device.wait_for_device = AsyncMock()
    with patch("builtins.open", mock_open()) as mock_file:
        mock_file.side_effect = IOError
        with patch(
            "mutenix.macropad.perform_upgrade_with_file",
        ) as mock_perform_upgrade:
            with pytest.raises(IOError):
                await macropad.manual_update(update_file)
            macropad._device.wait_for_device.assert_called_once()
            mock_file.assert_called_once_with(update_file, "rb")
            mock_perform_upgrade.assert_not_called()


@pytest.mark.asyncio
async def test_stop():
    macropad = Macropad()
    macropad._device = AsyncMock()
    macropad._websocket = AsyncMock()
    macropad._virtual_macropad = AsyncMock()

    await macropad.stop()

    macropad._device.stop.assert_called_once()
    macropad._websocket.stop.assert_called_once()
    macropad._virtual_macropad.stop.assert_called_once()


@pytest.mark.asyncio
async def test_update_device_status_teams_source_in_meeting(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.TEAMS,
            button_id=1,
            extra="is-muted",
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=True,
                isMuted=True,
                isHandRaised=False,
                isVideoOn=True,
            ),
            meetingPermissions=MeetingPermissions(canLeave=True),
        ),
    )

    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))
    macropad._virtual_macropad.send_msg.assert_called_once_with(SetLed(1, LedColor.RED))


@pytest.mark.asyncio
async def test_update_device_status_teams_source_not_in_meeting(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.TEAMS,
            button_id=1,
            extra="is-muted",
            color_on="red",
            color_off="green",
        ),
    ]
    macropad._current_state = ServerMessage(
        meetingUpdate=MeetingUpdate(
            meetingState=MeetingState(
                isInMeeting=False,
                isMuted=False,
                isHandRaised=False,
                isVideoOn=False,
            ),
            meetingPermissions=MeetingPermissions(canLeave=False),
        ),
    )

    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.BLACK))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.BLACK),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_with_result(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="echo blue",
            interval=0,
            read_result=True,
        ),
    ]
    macropad._last_status_check = defaultdict(int)
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    with patch("asyncio.to_thread", return_value="blue"):
        await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.BLUE))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.BLUE),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_without_result(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="exit 0",
            interval=0,
            read_result=False,
            color_on="yellow",
            color_off="black",
        ),
    ]
    macropad._last_status_check = defaultdict(int)
    macropad._device.send_msg = Mock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    with patch("asyncio.to_thread", return_value=0):
        await macropad._update_device_status()

    macropad._device.send_msg.assert_called_once_with(SetLed(1, LedColor.YELLOW))
    macropad._virtual_macropad.send_msg.assert_called_once_with(
        SetLed(1, LedColor.YELLOW),
    )


@pytest.mark.asyncio
async def test_update_device_status_cmd_source_interval_not_elapsed(macropad):
    macropad._config.leds = [
        Mock(
            source=LedStatusSource.CMD,
            button_id=1,
            extra="exit 0",
            interval=10,
            read_result=False,
            color_on="yellow",
            color_off="black",
        ),
    ]
    macropad._last_status_check = defaultdict(lambda: time.time())
    macropad._device.send_msg = AsyncMock()
    macropad._virtual_macropad.send_msg = AsyncMock()

    await macropad._update_device_status()

    macropad._device.send_msg.assert_not_called()
    macropad._virtual_macropad.send_msg.assert_not_called()
